/*
 * *   Copyright (C) 2017 Mukul S. Bansal (mukul.bansal@uconn.edu).
 * *
 * *   This program is free software: you can redistribute it and/or modify
 * *   it under the terms of the GNU General Public License as published by
 * *   the Free Software Foundation, either version 3 of the License, or
 * *   (at your option) any later version.
 * *
 * *   This program is distributed in the hope that it will be useful,
 * *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 * *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * *   GNU General Public License for more details.
 * *
 * *   You should have received a copy of the GNU General Public License
 * *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * */



#ifndef DTL_ALGORITHM_H
#define DTL_ALGORITHM_H

#include <set>
#include <algorithm>
#include <iterator>

class DTLreconciliation :  public TreeSet {
	
protected:
	vector< vector<int> > DParray; //create the dynamic programming table. (variable c(.,.) in paper).
	vector< vector<int> > in;
	vector< vector<int> > out;
	vector< vector<int> > in_second; // for use with the out counter (can't use the regular in() counters because they count losses)
	int c_spec, c_dup, c_trans, minIn, minIn_second, minOut, minFinal, index, index2, numDup, numTrans;
	int totalCost, totalDup, totalTrans, totalLoss;
	int numOptRootings; // to store the total number of optimal rootings of the gene tree.

	
	GeneNodeRooted *bestRoot; // to store the best rooting for unrooted gene trees.
	int bestScore; // to store the bestScore found so far when trying alternative roots for the gene tree.
	struct rootdata     // struct for the queue for storing optimal rootings
	{
		GeneNodeRooted *optroot;
	};
	
	rootdata temproot;
	vector <rootdata> bestRootQueue;
	
	struct data{
		SpeciesNode *mapping0; // left child's mapping
		SpeciesNode *mapping1; // right child's mapping
		SpeciesNode *recipient;  // to store the recipient in case of transfer
		int event;		// 0 = speciation, 1 = duplication, 2 = transfer
		
	};
	
	vector< vector<data> > backTrack;    // table for back tracking
	vector<data> optQueue;  //create the temporary structure that will store all equally optimal scenarios encountered during nested post order

	
	// structure for storing the optimal mapping for the in, out, and in_second counters.
	struct datainout{
		SpeciesNode *optmapping; // using a struct for this simply for extensibility
	};
	
	vector< vector<datainout> > inTrack;   // tables for storing optimal mappings for the in, out, and in_second counters.
	vector< vector<datainout> > outTrack;
	vector< vector<datainout> > in_secondTrack; 		
	
	
	data temp; // create temporary instance of structure for storing intermediate optimal mappings and event type.
	datainout tempinout; // create temporary instance of inout structure for storing intermediate optimal mapping.	
	
	
public:
	DTLreconciliation() {
		
	}

	
/*
	void SetGeneTreePOidx(GeneNodeRooted *currnode, int index) 
	{
		if (currnode == NULL) return;
		SetGeneTreePOidx(currnode->child(0), index);
		SetGeneTreePOidx(currnode->child(1), index);
		
		cout << currnode->idx << endl;
	}
	
	
	void SetSpeciesTreePOidx(SpeciesNode *currnode, int index) 
	{
		if (currnode == NULL) return;
		SetSpeciesTreePOidx(currnode->child(0), index);
		SetSpeciesTreePOidx(currnode->child(1), index);
		
		cout << currnode->idx << endl;

	}
	
*/	
	
	// set the leftmost decendant and rightmost descendants for the nodes of the gene tree.
	void leftRightDescendants(GeneNodeRooted *currnode) 
	{
		
		if (currnode == NULL) return;
		
		leftRightDescendants(currnode->child(0));
		leftRightDescendants(currnode->child(1));
		
		if (currnode->isLeaf() == true)
		{
			currnode->leftMost = currnode;
			currnode->rightMost = currnode;
			return;
		}
		else
		{
			currnode->leftMost = currnode->child(0)->leftMost;
			currnode->rightMost = currnode->child(1)->rightMost;
		}
		
	}
	
	
	
		// do a preOrder traversal of the species tree to label nodes with their depth (used for computing distances between nodes)
	void doNodePreOrder(SpeciesNode *node, int depthCount) {

			node->nodeDepth = depthCount;
			depthCount++;
		
		if (node->child(0) != NULL) {
			doNodePreOrder(node->child(0), depthCount);
		}
		if (node->child(1) != NULL) {
			doNodePreOrder(node->child(1), depthCount);
		}
	}

	
	
	
	
	
	// do a preOrder traversal of the species tree to label nodes with their depth in terms of branch lengths (for computing transfers)
	void doBranchPreOrder(SpeciesNode *node) {
		if (node == speciestree->root)
		{
			node->branchDepth = node->weight;
		}
		else
		{
			node->branchDepth = node->weight + node->parent()->branchDepth;
		}
		
		if (node->child(0) != NULL) {
			doBranchPreOrder(node->child(0));
		}
		if (node->child(1) != NULL) {
			doBranchPreOrder(node->child(1));
		}
	}

	
	
	// Perform nested post order traversal (main part of DTL-recociliation algo)
	void PostOrderGeneTree(GeneNodeRooted *currnode) 
	{
		if (currnode == NULL) return;
		PostOrderGeneTree(currnode->child(0));
		PostOrderGeneTree(currnode->child(1));
		
		// perform post order traversal of the species tree
		PostOrderSpeciesTree(currnode, speciestree->root);
		
		// update the out[][] counters as well as the outTrack table 
//		PreOrderSpeciesTree(currnode, speciestree->root);
		
		
	}
	
	
	void PostOrderSpeciesTree(GeneNodeRooted *currnode, SpeciesNode *CurrSpeciesNode) 
	{


		if (CurrSpeciesNode == NULL) return;
		PostOrderSpeciesTree(currnode, CurrSpeciesNode->child(0));
		PostOrderSpeciesTree(currnode, CurrSpeciesNode->child(1));

		
		// build up the DP table
		if(currnode->isLeaf() == false)
		{
			
			optQueue.clear(); // clear the queue that stores all optimal solutions
//			data temp; // create temporary instance of structure for storing optimal mappings and event type.
//			datainout tempinout; // create temporary instance of inout structure for storing optimal mapping.
			
			if(CurrSpeciesNode->isLeaf() == true)
			{
				
				
				// We always divide INT_MAX by 10 just to prevent overflow of the int value when we are adding two large integers.
				
				minFinal = INT_MAX / 10;  // to keep track of min value among speciation, duplication and transfer.
				
				//speciation 
				c_spec = INT_MAX / 10;
				
				
				// duplication 
				c_dup = DCOST + DParray[currnode->child(0)->idx][CurrSpeciesNode->idx] + DParray[currnode->child(1)->idx][CurrSpeciesNode->idx];
				
				if (c_dup < minFinal) 
				{
					minFinal = c_dup;
					optQueue.clear();
					temp.mapping0 = CurrSpeciesNode;
					temp.mapping1 = CurrSpeciesNode;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				else if ((c_dup == minFinal) || (minFinal >= INT_MAX / 10)) 
				{
					temp.mapping0 = CurrSpeciesNode;
					temp.mapping1 = CurrSpeciesNode;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				
				
				
				// transfer
				c_trans = INT_MAX / 10;
				int T1, T2;
				if(CurrSpeciesNode != speciestree->root)
				{
					// first transfer scenario
					T1 = in[currnode->child(0)->idx][CurrSpeciesNode->idx] + getBestOut(currnode->child(1), CurrSpeciesNode);
					
					c_trans = TCOST + T1;
				
					if (c_trans < minFinal) 
					{
						minFinal = c_trans;
						optQueue.clear();
						temp.mapping0 = inTrack[currnode->child(0)->idx][CurrSpeciesNode->idx].optmapping;
						temp.mapping1 = outTrack[currnode->child(1)->idx][CurrSpeciesNode->idx].optmapping;
						temp.recipient = temp.mapping1;
						temp.event = 2;
						optQueue.push_back(temp);
					}
					else if ((c_trans == minFinal) || (minFinal >= INT_MAX / 10)) 
					{
						temp.mapping0 = inTrack[currnode->child(0)->idx][CurrSpeciesNode->idx].optmapping;
						temp.mapping1 = outTrack[currnode->child(1)->idx][CurrSpeciesNode->idx].optmapping;
						temp.recipient = temp.mapping1;
						temp.event = 2;
						optQueue.push_back(temp);
					}
					
					
					
					// second transfer scenario
					T2 = in[currnode->child(1)->idx][CurrSpeciesNode->idx] + getBestOut(currnode->child(0), CurrSpeciesNode);
					
					c_trans = TCOST + T2;
					
					if (c_trans < minFinal) 
					{
						minFinal = c_trans;
						optQueue.clear();
						temp.mapping0 = outTrack[currnode->child(0)->idx][CurrSpeciesNode->idx].optmapping;
						temp.mapping1 = inTrack[currnode->child(1)->idx][CurrSpeciesNode->idx].optmapping;
						temp.recipient = temp.mapping0;
						temp.event = 2;
						optQueue.push_back(temp);
					}
					else if ((c_trans == minFinal) || (minFinal >= INT_MAX / 10)) 
					{
						temp.mapping0 = outTrack[currnode->child(0)->idx][CurrSpeciesNode->idx].optmapping;
						temp.mapping1 = inTrack[currnode->child(1)->idx][CurrSpeciesNode->idx].optmapping;
						temp.recipient = temp.mapping0;
						temp.event = 2;
						optQueue.push_back(temp);
					}

				}
				
				
				
				// pick a randon optimal solution from the queue
				if (!optQueue.empty()) 
				{
					index = rand() % optQueue.size();
				}
				else
				{
					cout << "ERROR!! Empty Queue." << endl;
					exit(1);
				}	
				
				
				// set DParray entry 
				DParray[currnode->idx][CurrSpeciesNode->idx] = minFinal;
				
				// update the backtracking array
				backTrack[currnode->idx][CurrSpeciesNode->idx].mapping0 =  optQueue[index].mapping0;
				backTrack[currnode->idx][CurrSpeciesNode->idx].mapping1 =  optQueue[index].mapping1;
				backTrack[currnode->idx][CurrSpeciesNode->idx].recipient =  optQueue[index].recipient;
				backTrack[currnode->idx][CurrSpeciesNode->idx].event =  optQueue[index].event;
				
				
				
				// update in[][] entry as well as inTrack
				in[currnode->idx][CurrSpeciesNode->idx] = DParray[currnode->idx][CurrSpeciesNode->idx];
				inTrack[currnode->idx][CurrSpeciesNode->idx].optmapping = CurrSpeciesNode;
				
				// update in_second[][] entry as well as in_secondTrack
				in_second[currnode->idx][CurrSpeciesNode->idx] = DParray[currnode->idx][CurrSpeciesNode->idx];
				in_secondTrack[currnode->idx][CurrSpeciesNode->idx].optmapping = CurrSpeciesNode;
				
				
/*				
if((currnode->idx ==1) && (CurrSpeciesNode->idx == 1))
cout << endl << "BEGIN " << DParray[currnode->child(0)->idx][CurrSpeciesNode->idx] << "  " <<  out[currnode->child(1)->idx][CurrSpeciesNode->idx] << " " << DParray[currnode->idx][CurrSpeciesNode->idx];
*/				
			}
			else
			{
				
				minFinal = INT_MAX / 10;  // to keep track of min value among speciation, duplication and transfer.
				
				// speciation
				int S1 = in[currnode->child(0)->idx][CurrSpeciesNode->child(0)->idx] + in[currnode->child(1)->idx][CurrSpeciesNode->child(1)->idx];
				
				c_spec = S1;
				
				if (c_spec < minFinal) 
				{
					minFinal = c_spec;
					optQueue.clear();
					temp.mapping0 = inTrack[currnode->child(0)->idx][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.mapping1 = inTrack[currnode->child(1)->idx][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 0;
					optQueue.push_back(temp);
				}
				else if ((c_spec == minFinal) || (minFinal >= INT_MAX / 10)) 
				{
					temp.mapping0 = inTrack[currnode->child(0)->idx][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.mapping1 = inTrack[currnode->child(1)->idx][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 0;
					optQueue.push_back(temp);
				}
				
				
				
				int S2 = in[currnode->child(0)->idx][CurrSpeciesNode->child(1)->idx] + in[currnode->child(1)->idx][CurrSpeciesNode->child(0)->idx];

				c_spec = S2;
				
				if (c_spec < minFinal) 
				{
					minFinal = c_spec;
					optQueue.clear();
					temp.mapping0 = inTrack[currnode->child(0)->idx][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.mapping1 = inTrack[currnode->child(1)->idx][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 0;
					optQueue.push_back(temp);
				}
				else if ((c_spec == minFinal) || (minFinal >= INT_MAX / 10)) 
				{
					temp.mapping0 = inTrack[currnode->child(0)->idx][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.mapping1 = inTrack[currnode->child(1)->idx][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 0;
					optQueue.push_back(temp);
				}
				
				

				
				
				
				// duplication
//				minD = INT_MAX / 10;
				int D1 = LCOST + DParray[currnode->child(0)->idx][CurrSpeciesNode->idx] + in[currnode->child(1)->idx][CurrSpeciesNode->child(1)->idx];
					
				c_dup = DCOST + D1;
								
				if (c_dup < minFinal) 
				{
					minFinal = c_dup;
					optQueue.clear();
					temp.mapping0 = CurrSpeciesNode;
					temp.mapping1 = inTrack[currnode->child(1)->idx][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				else if ((c_dup == minFinal) || (minFinal >= INT_MAX / 10)) 
				{
					temp.mapping0 = CurrSpeciesNode;
					temp.mapping1 = inTrack[currnode->child(1)->idx][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}

				
				int D2 = LCOST + DParray[currnode->child(0)->idx][CurrSpeciesNode->idx] + in[currnode->child(1)->idx][CurrSpeciesNode->child(0)->idx];
				
				c_dup = DCOST + D2;
								
				if (c_dup < minFinal) 
				{
					minFinal = c_dup;
					optQueue.clear();
					temp.mapping0 = CurrSpeciesNode;
					temp.mapping1 = inTrack[currnode->child(1)->idx][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				else if ((c_dup == minFinal) || (minFinal >= INT_MAX / 10)) 
				{
					temp.mapping0 = CurrSpeciesNode;
					temp.mapping1 = inTrack[currnode->child(1)->idx][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				
				int D3 = LCOST + DParray[currnode->child(1)->idx][CurrSpeciesNode->idx] + in[currnode->child(0)->idx][CurrSpeciesNode->child(1)->idx];
				
				c_dup = DCOST + D3;
								
				if (c_dup < minFinal) 
				{
					minFinal = c_dup;
					optQueue.clear();
					temp.mapping0 = inTrack[currnode->child(0)->idx][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.mapping1 = CurrSpeciesNode;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				else if ((c_dup == minFinal) || (minFinal >= INT_MAX / 10)) 
				{
					temp.mapping0 = inTrack[currnode->child(0)->idx][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.mapping1 = CurrSpeciesNode;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				
				int D4 = LCOST + DParray[currnode->child(1)->idx][CurrSpeciesNode->idx] + in[currnode->child(0)->idx][CurrSpeciesNode->child(0)->idx];
				
				c_dup = DCOST + D4;
								
				if (c_dup < minFinal) 
				{
					minFinal = c_dup;
					optQueue.clear();
					temp.mapping0 = inTrack[currnode->child(0)->idx][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.mapping1 = CurrSpeciesNode;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				else if ((c_dup == minFinal) || (minFinal >= INT_MAX / 10)) 
				{
					temp.mapping0 = inTrack[currnode->child(0)->idx][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.mapping1 = CurrSpeciesNode;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				
				int D5 = DParray[currnode->child(0)->idx][CurrSpeciesNode->idx] + DParray[currnode->child(1)->idx][CurrSpeciesNode->idx];
				
				c_dup = DCOST + D5;
								
				if (c_dup < minFinal) 
				{
					minFinal = c_dup;
					optQueue.clear();
					temp.mapping0 = CurrSpeciesNode;
					temp.mapping1 = CurrSpeciesNode;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				else if ((c_dup == minFinal) || (minFinal >= INT_MAX / 10)) 
				{
					temp.mapping0 = CurrSpeciesNode;
					temp.mapping1 = CurrSpeciesNode;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				
				
				int D6 = (2 * LCOST) + in[currnode->child(0)->idx][CurrSpeciesNode->child(0)->idx] + in[currnode->child(1)->idx][CurrSpeciesNode->child(1)->idx];
				
				c_dup = DCOST + D6;
								
				if (c_dup < minFinal) 
				{
					minFinal = c_dup;
					optQueue.clear();
					temp.mapping0 = inTrack[currnode->child(0)->idx][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.mapping1 = inTrack[currnode->child(1)->idx][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				else if ((c_dup == minFinal) || (minFinal >= INT_MAX / 10)) 
				{
					temp.mapping0 = inTrack[currnode->child(0)->idx][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.mapping1 = inTrack[currnode->child(1)->idx][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				
				int D7 = (2 * LCOST) + in[currnode->child(0)->idx][CurrSpeciesNode->child(1)->idx] + in[currnode->child(1)->idx][CurrSpeciesNode->child(0)->idx];
				
				c_dup = DCOST + D7;
								
				if (c_dup < minFinal) 
				{
					minFinal = c_dup;
					optQueue.clear();
					temp.mapping0 = inTrack[currnode->child(0)->idx][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.mapping1 = inTrack[currnode->child(1)->idx][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				else if ((c_dup == minFinal) || (minFinal >= INT_MAX / 10)) 
				{
					temp.mapping0 = inTrack[currnode->child(0)->idx][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.mapping1 = inTrack[currnode->child(1)->idx][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				
				int D8 = (2 * LCOST) + in[currnode->child(0)->idx][CurrSpeciesNode->child(0)->idx] + in[currnode->child(1)->idx][CurrSpeciesNode->child(0)->idx];
				
				c_dup = DCOST + D8;
								
				if (c_dup < minFinal) 
				{
					minFinal = c_dup;
					optQueue.clear();
					temp.mapping0 = inTrack[currnode->child(0)->idx][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.mapping1 = inTrack[currnode->child(1)->idx][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				else if ((c_dup == minFinal) || (minFinal >= INT_MAX / 10)) 
				{
					temp.mapping0 = inTrack[currnode->child(0)->idx][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.mapping1 = inTrack[currnode->child(1)->idx][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				
				int D9 = (2 * LCOST) + in[currnode->child(0)->idx][CurrSpeciesNode->child(1)->idx] + in[currnode->child(1)->idx][CurrSpeciesNode->child(1)->idx];
				
				c_dup = DCOST + D9;
								
				if (c_dup < minFinal) 
				{
					minFinal = c_dup;
					optQueue.clear();
					temp.mapping0 = inTrack[currnode->child(0)->idx][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.mapping1 = inTrack[currnode->child(1)->idx][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				else if ((c_dup == minFinal) || (minFinal >= INT_MAX / 10)) 
				{
					temp.mapping0 = inTrack[currnode->child(0)->idx][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.mapping1 = inTrack[currnode->child(1)->idx][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				

				
				
				// transfer
				c_trans = INT_MAX / 10; 
				if(CurrSpeciesNode != speciestree->root)
				{
					int T1 = in[currnode->child(0)->idx][CurrSpeciesNode->idx] + getBestOut(currnode->child(1), CurrSpeciesNode);
					
					
					c_trans = TCOST + T1;
					
					if (c_trans < minFinal) 
					{
						minFinal = c_trans;
						optQueue.clear();
						temp.mapping0 = inTrack[currnode->child(0)->idx][CurrSpeciesNode->idx].optmapping;
						temp.mapping1 = outTrack[currnode->child(1)->idx][CurrSpeciesNode->idx].optmapping;
						temp.recipient = temp.mapping1;
						temp.event = 2;
						optQueue.push_back(temp);
					}
					else if ((c_trans == minFinal) || (minFinal >= INT_MAX / 10)) 
					{
						temp.mapping0 = inTrack[currnode->child(0)->idx][CurrSpeciesNode->idx].optmapping;
						temp.mapping1 = outTrack[currnode->child(1)->idx][CurrSpeciesNode->idx].optmapping;
						temp.recipient = temp.mapping1;
						temp.event = 2;
						optQueue.push_back(temp);
					}
					
					
					int T2 = in[currnode->child(1)->idx][CurrSpeciesNode->idx] + getBestOut(currnode->child(0), CurrSpeciesNode);
					
					c_trans = TCOST + T2;
					
					if (c_trans < minFinal) 
					{
						minFinal = c_trans;
						optQueue.clear();
						temp.mapping0 = outTrack[currnode->child(0)->idx][CurrSpeciesNode->idx].optmapping;
						temp.mapping1 = inTrack[currnode->child(1)->idx][CurrSpeciesNode->idx].optmapping;
						temp.recipient = temp.mapping0;
						temp.event = 2;
						optQueue.push_back(temp);
					}
					else if ((c_trans == minFinal) || (minFinal >= INT_MAX / 10)) 
					{
						temp.mapping0 = outTrack[currnode->child(0)->idx][CurrSpeciesNode->idx].optmapping;
						temp.mapping1 = inTrack[currnode->child(1)->idx][CurrSpeciesNode->idx].optmapping;
						temp.recipient = temp.mapping0;
						temp.event = 2;
						optQueue.push_back(temp);
					}
					
				}
					
				
				
				// pick a random optimal solution from the queue
				if (!optQueue.empty()) 
				{
					index = rand() % optQueue.size();
				}
				else
				{
					cout << "ERROR!! Empty Queue." << endl;
					exit(1);
				}
				
				
				// set DParray entry
				DParray[currnode->idx][CurrSpeciesNode->idx] = minFinal;
				
				// update the backtracking array
				backTrack[currnode->idx][CurrSpeciesNode->idx].mapping0 =  optQueue[index].mapping0;
				backTrack[currnode->idx][CurrSpeciesNode->idx].mapping1 =  optQueue[index].mapping1;
				backTrack[currnode->idx][CurrSpeciesNode->idx].recipient =  optQueue[index].recipient;
				backTrack[currnode->idx][CurrSpeciesNode->idx].event =  optQueue[index].event;
				
				
				// update in[][] entry and the inTrack table
				
				minIn = DParray[currnode->idx][CurrSpeciesNode->idx];
				tempinout.optmapping = CurrSpeciesNode; 
				if ( (LCOST + in[currnode->idx][CurrSpeciesNode->child(0)->idx]) < minIn)
				{
					minIn = (LCOST + in[currnode->idx][CurrSpeciesNode->child(0)->idx]);
					tempinout.optmapping = inTrack[currnode->idx][CurrSpeciesNode->child(0)->idx].optmapping;
				}
				if ( (LCOST + in[currnode->idx][CurrSpeciesNode->child(1)->idx]) < minIn)
				{
					minIn = (LCOST + in[currnode->idx][CurrSpeciesNode->child(1)->idx]);
					tempinout.optmapping = inTrack[currnode->idx][CurrSpeciesNode->child(1)->idx].optmapping;
				}
				
				in[currnode->idx][CurrSpeciesNode->idx] = minIn;
				inTrack[currnode->idx][CurrSpeciesNode->idx].optmapping = tempinout.optmapping;
				
			
				
				
				// update in_second[][] entry and the in_secondTrack table
				
				minIn_second = DParray[currnode->idx][CurrSpeciesNode->idx];
				tempinout.optmapping = CurrSpeciesNode; 
				if ( in_second[currnode->idx][CurrSpeciesNode->child(0)->idx] < minIn_second)
				{
					minIn_second = in_second[currnode->idx][CurrSpeciesNode->child(0)->idx];
					tempinout.optmapping = in_secondTrack[currnode->idx][CurrSpeciesNode->child(0)->idx].optmapping;
				}
				if ( in_second[currnode->idx][CurrSpeciesNode->child(1)->idx] < minIn_second)
				{
					minIn_second = in_second[currnode->idx][CurrSpeciesNode->child(1)->idx];
					tempinout.optmapping = in_secondTrack[currnode->idx][CurrSpeciesNode->child(1)->idx].optmapping;
				}
				
				in_second[currnode->idx][CurrSpeciesNode->idx] = minIn_second;
				in_secondTrack[currnode->idx][CurrSpeciesNode->idx].optmapping = tempinout.optmapping;
			
			}
			
		}
	}
	
	
	
		// This function computes the best receiver for a transfer (by brute force to account for distance dependant transfer costs).
	int getBestOut(GeneNodeRooted *currnode, SpeciesNode *node)
	{
		minOut = INT_MAX / 10;
			
		bool valid;
		int transferAdd;   // additional cost of transfer based on distance
		int num_nodes = speciestree->nodes.size();
		int distance;
		SpeciesNode* temp_node;
		SpeciesNode* lcanode;
		
		// iterate over all non-root nodes
		for (int i = 1; i< num_nodes; i++)
		{
			
//if (speciestree->nodes[i]->child(0) == NULL)
//cout << speciestree->nodes[i]->branchDepth << endl;
			
                       if (speciestree->nodes[i] == node)
                                continue;


                        valid = true;

                        // disallow transfers that go from descendant to ancestor or vice versa. All other transfers are OK.
                        if ((node->no >= speciestree->nodes[i]->begin) && (node->no <= speciestree->nodes[i]->end))
                                valid = false;
                        if ((speciestree->nodes[i]->no >= node->begin) && (speciestree->nodes[i]->no <= node->end))
                                valid = false;

			
			
			if (valid == true)
			{
				
				temp_node = node;
				while(temp_node->parent() != NULL)
				{
					temp_node = temp_node->parent();
					if((speciestree->nodes[i]->no >= temp_node->begin) && (speciestree->nodes[i]->no <= temp_node->end))
						break;
				}
				
				lcanode = temp_node;
				
				
				distance = (node->nodeDepth - lcanode->nodeDepth) + (speciestree->nodes[i]->nodeDepth - lcanode->nodeDepth);
				
				
				
				
				
//				if((node->no >= speciestree->nodes[i]->begin) && (node->no <= speciestree->nodes[i]->end))
//					cout << "ERROR!!!" << endl;
//				if((speciestree->nodes[i]->no >= node->begin) && (speciestree->nodes[i]->no <= node->end))
//					cout << "ERROR!!!" << endl;
				
				
				if (TYPE == 0)   // if not using variable transfer costs
				{
					if (DParray[currnode->idx][speciestree->nodes[i]->idx] < minOut)
					{
						minOut = DParray[currnode->idx][speciestree->nodes[i]->idx];
						outTrack[currnode->idx][node->idx].optmapping = speciestree->nodes[i];
					}
				}
				if (TYPE == 1)  // if using the simple transfer cost
				{
					
					if (distance >= THRESHOLD)
						transferAdd = ADD;
					else
						transferAdd = 0;
					
					if ((DParray[currnode->idx][speciestree->nodes[i]->idx] + transferAdd) < minOut)
					{
						minOut = (DParray[currnode->idx][speciestree->nodes[i]->idx] + transferAdd);
						outTrack[currnode->idx][node->idx].optmapping = speciestree->nodes[i];
					}
					
				}
				
				
				if (TYPE == 2)   // if using the more general transfer cost
				{
					
					if (distance >= THRESHOLD)
						transferAdd = (distance/THRESHOLD) * ADD;
					else
						transferAdd = 0;
					
					if ((DParray[currnode->idx][speciestree->nodes[i]->idx] + transferAdd) < minOut)
					{
						minOut = (DParray[currnode->idx][speciestree->nodes[i]->idx] + transferAdd);
						outTrack[currnode->idx][node->idx].optmapping = speciestree->nodes[i];
					}
				}
				
				

				
				
			}

		}
 		
		return minOut;
	}
	
	
	
	
	
	
	
	// output the reconciliation cost and call function for outputting actual reconciliation
	void OutputReconciliationCost(ostream &os, GeneNodeRooted *currnode, bool random)
	{
		
		int minReconCost = INT_MAX / 10;
		int numSpecies_nodes = speciestree->nodes.size();
		int bestMapIdx;
		
		for (int i = 0; i < numSpecies_nodes; i++)
		{
			if (DParray[currnode->idx][i] < minReconCost)
			{
				minReconCost = DParray[currnode->idx][i];
				bestMapIdx = i;
			}
		}
		
		
			// call function to output the actual reconciliation 
			
			os << endl << endl << "The minimum reconciliation cost is: " << minReconCost;
			os << endl << "The total number of optimal rootings is: " << numOptRootings << endl;

			
				

	}
	
	
		
	
/*	
	// return the reconciliation cost (for use with evaluating different rootings for unrooted gene trees)
	int returnReconciliationCost(GeneNodeRooted *currnode)
	{
		
		int minReconCost = INT_MAX / 10;
		int numSpecies_nodes = speciestree->nodes.size();
		
		for (int i = 0; i < numSpecies_nodes; i++)
		{
			if (DParray[currnode->idx][i] < minReconCost)
			{
				minReconCost = DParray[currnode->idx][i];
			}
		}
		
		return minReconCost;	

	}
	
*/
	
	void run(ostream &os, const ReRoot reroot, bool random) 
	{
		bool rerooting = (reroot == ALL);
		createLeafMapping();
		speciestree->establishOrder();
		speciestree->assignIndex();
		
		
		// initialize the counters that store the total cost and number of events across all gene trees for the summary statistics.
		totalCost = 0;
		totalDup = 0;
		totalTrans = 0;
		totalLoss = 0;


//		set<int> genetreecount;
//		genetreecount.clear();
		
		msgout << "Computing...\n";


		doNodePreOrder(speciestree->root, 1);
		doBranchPreOrder(speciestree->root);
		
//SetSpeciesTreePOidx(speciestree->root, 0);
		
		// call the nested post-order procedure to solve the DTL-problem
		for(int i2=0; i2<genetree_rooted.size(); i2++) 
		{
			GeneTreeRooted &tree = *genetree_rooted[i2];
			tree.assignIndex();
			
			// initialize the numDup and numTrans counters (to store the final number of duplications and transfers in the reconciliation)
			numDup = 0;
			numTrans = 0;
			
			// Initialize the dynamic programming table and in and out counters to infinity
			
			int numGene_nodes = tree.nodes.size();
			int numSpecies_nodes = speciestree->nodes.size();
			
			vector<int> row;
			row.clear();
			
			vector<data> datarow;
			datarow.clear();
			data temp1;
			vector<datainout> datainoutrow;
			datainoutrow.clear();
			datainout temp2;
			
			
			for (int i = 0; i< numGene_nodes; i++)
			{		
				
				for (int j = 0; j< numSpecies_nodes;j++)
				{
					row.push_back(INT_MAX / 10);
					
					temp1.mapping0 = NULL;
					temp1.mapping1 = NULL;
					temp1.recipient = NULL;
					temp1.event = -1;
					datarow.push_back(temp1);
					
					temp2.optmapping = NULL;
					datainoutrow.push_back(temp2);
					
				}
				
				DParray.push_back(row);
				in.push_back(row);
				out.push_back(row);
				in_second.push_back(row);
				
				backTrack.push_back(datarow);
				inTrack.push_back(datainoutrow);
				outTrack.push_back(datainoutrow);
				in_secondTrack.push_back(datainoutrow);
				
				row.clear();
				datarow.clear();
				datainoutrow.clear();
			}
			
			
			
			
			// base case of the dynamic programming table (incorporate leaf to leaf mapping)	
			vector<GeneNodeRooted*> &g = tree.nodes;
			SpeciesNode *temp_node;

		

			for(vector<GeneNodeRooted*>::iterator itr=g.begin(); itr!=g.end(); itr++) 
			{
				GeneNodeRooted *genenode = *itr;
				if (genenode->isLeaf())
				{
					SpeciesNode *&mapping = genenode->getMapping();
					DParray[genenode->idx][mapping->idx] = 0;
					
					temp_node = genenode->getMapping();
					int count = 0;
					while (1)
					{
						in[genenode->idx][temp_node->idx] = (count * LCOST);
						in_second[genenode->idx][temp_node->idx] = 0;
						
						inTrack[genenode->idx][temp_node->idx].optmapping = genenode->getMapping();
						in_secondTrack[genenode->idx][temp_node->idx].optmapping = genenode->getMapping();
						
						if (temp_node == speciestree->root)
							break;
						temp_node = temp_node->parent();
						count++;
						
					}
					
				}
			}
		
			
			if (tree.rooting == true)    // if the gene tree is rooted
			{
			
				// execute the main nested post order traversal
				PostOrderGeneTree(tree.root);
			
				
					// Output the species tree and gene tree
					os << endl << endl << " ------------ Optimal Rootings for Gene Tree " << i2+1 << " (rooted) -------------" << endl;
//					os << "Species Tree: " << endl; 
//					speciestree->tree2newick(os);
//					os << endl << endl;
//					os << "Gene Tree: " << endl;
					tree.genetree2newick(os);
					os << endl;
//	cout << tree << endl;
			
			
				// decorate the nodes of the gene tree with their leftmost and rightmost descendants.
				leftRightDescendants(tree.root);
			
				// output the final reconciliation cost
				OutputReconciliationCost(os,tree.root, random);
				

			}
			else  // if the gene tree is unrooted
			{
				// execute the main nested post order traversal for the original rooting of the gene tree
				PostOrderGeneTree(tree.root);
				
				bestRoot = tree.root;
				bestRootQueue.clear();
				temproot.optroot = bestRoot;
				bestRootQueue.push_back(temproot);
//cout << "reached here 1? " << bestScore << endl; 
//cout << tree << endl;
				
				bestScore = outputCurrentCost();
				
				numOptRootings = 1;
				eulerTour(tree.root, tree.root);   // call the main euler tour function for the corresponding gene tree and find the best rooting
				//cout << "reached here 2? " << bestScore << endl; 
				
				// pick an optimal rooting at random
                               if (random == 1)
                                {
                                        if (!bestRootQueue.empty())
                                        {
                                                index2 = rand() % bestRootQueue.size();
                                                tree.reroot(bestRootQueue[index2].optroot);   // reroot the gene tree to its best rooting.

                                                // compute the reconciliation for the newly rerooted tree.
                                                runForNewRoot(i2, os, random);
                                        }
                                        else
                                        {
                                                cout << "ERROR: bestRootQueue is empty!!" << endl;
                                                exit(1);
                                        }
                                }
                                else			
				 {

                                        if (!bestRootQueue.empty())
                                        {
                                                GeneNodeRooted *rootsleftchild = tree.root->child(0);
                                                GeneNodeRooted *rootsrightchild = tree.root->child(1);
                                                os << endl << endl << " ------------ All Optimal Rootings for Gene Tree " << i2+1 << " (Unrooted) -------------" << endl;
                                                for (index2 = 0; index2 < bestRootQueue.size(); index2++)
                                                {
                                                        tree.reroot(bestRootQueue[index2].optroot);      // reroot the gene tree to its best rooting.
                                                        // compute the reconciliation for the newly rerooted tree.

                                                        tree.genetree2newick(os);
                                                        os << endl;

                                                        // restore the root to its original position so the next reroot operation can be performed properly.
                                                        if (rootsleftchild->parent() == rootsrightchild)
                                                                tree.reroot(rootsleftchild);
                                                        else
                                                                tree.reroot(rootsrightchild);

                                                }
                                                // compute the reconciliation for the optimally rerooted trees.
                                                tree.reroot(bestRootQueue[0].optroot);
                                                runForNewRoot(i2, os, random);

                                        }
                                }

			
				
			}
			
			
			
/*			
			
			
			
cout << endl<< "DParray " << DParray.size() << " " << DParray[0].size() << endl;
for (int i =0 ; i< DParray.size(); i++)
{
for (int j=0; j < DParray[0].size();j++)
{
cout << DParray[i][j] << " ";
}
cout << endl;			
}
			

			
			
			
cout << endl << "IN" << endl;			
//cout << endl << in.size() << " " << in[0].size() << endl;
for (int i =0 ; i< in.size(); i++)
{
for (int j=0; j < in[0].size();j++)
{
cout << in[i][j] << " ";
}
cout << endl;			
}
		
			

cout << endl << "IN TRACK" << endl;			
//cout << endl << in.size() << " " << in[0].size() << endl;
for (int i =0 ; i< inTrack.size(); i++)
{
for (int j=0; j < inTrack[0].size();j++)
{
if(	inTrack[i][j].optmapping == NULL)
cout << "NULL ";
else
cout << ((NamedSpeciesNode*)inTrack[i][j].optmapping)->getName() << " ";
}
cout << endl;			
}			
			
			
			
			
cout << endl << "IN_SECOND" << endl;			
//cout << endl << in.size() << " " << in[0].size() << endl;
for (int i =0 ; i< in_second.size(); i++)
{
for (int j=0; j < in_second[0].size();j++)
{
cout << in_second[i][j] << " ";
}
cout << endl;			
}


cout << endl << "IN_SECOND TRACK" << endl;			
//cout << endl << in.size() << " " << in[0].size() << endl;
for (int i =0 ; i< in_secondTrack.size(); i++)
{
for (int j=0; j < in_secondTrack[0].size();j++)
{
if(	in_secondTrack[i][j].optmapping == NULL)
cout << "NULL ";
else
cout << ((NamedSpeciesNode*)in_secondTrack[i][j].optmapping)->getName() << " ";
}
cout << endl;			
}
			
			
			
cout << endl << "OUT" << endl;			
//cout << endl << in.size() << " " << in[0].size() << endl;
for (int i =0 ; i< out.size(); i++)
{
for (int j=0; j < out[0].size();j++)
{
cout << out[i][j] << " ";
}
cout << endl;			
}
			


cout << endl << "OUT TRACK" << endl;			
//cout << endl << in.size() << " " << in[0].size() << endl;
for (int i =0 ; i< outTrack.size(); i++)
{
for (int j=0; j < outTrack[0].size();j++)
{
if(	outTrack[i][j].optmapping == NULL)
cout << "NULL ";
else
cout << ((NamedSpeciesNode*)outTrack[i][j].optmapping)->getName() << " ";
}
cout << endl;			
}
			
			
			
*/			
			
			
			
			
		
			// clean up the arrays
			for (int i =0 ; i< DParray.size(); i++)
			{
				DParray[i].clear();
				in[i].clear();
				in_second[i].clear();
				out[i].clear();
				
				backTrack[i].clear();
				inTrack[i].clear();
				outTrack[i].clear();
				in_secondTrack[i].clear();
				
			}
			DParray.clear();
			in.clear();
			in_second.clear();
			out.clear();
			
			backTrack.clear();
			inTrack.clear();
			outTrack.clear();
			in_secondTrack.clear();

			
		}
		
		
		
		

//	speciestree->tree2newick(os);
		
//	cout << endl << *speciestree << endl;
				
			
	}

	
	
	// perform the euler tour for unrooted gene trees to find the best rooting.
	void eulerTour(GeneNodeRooted *currnode, GeneNodeRooted *treeroot)
	{

		
//cout << currnode << " " << currnode->idx << endl;
		
		// skip certain nodes
		if (currnode->idx != 0)
		{	
			if (currnode->parent()->idx != 0)
			{
				
			if(currnode->parent()->parent()->idx == 0)   // special case when the grandparent of current node is the root of the original gene tree
			{
				// if moving down to the left-child (i.e., if the current node is the left child of its parent) 
				if(currnode == currnode->parent()->child(0))
				{	
					// call the functions for updating the tables for the two nodes that need to be updated (parent and grandparent of the current node)
					PostOrderSpeciesTreeReroot(currnode->parent(), currnode->getSibling()->idx, currnode->parent()->getSibling()->idx, speciestree->root);
//					PreOrderSpeciesTree(currnode->parent(), speciestree->root);
					
					PostOrderSpeciesTreeReroot(treeroot, currnode->idx, currnode->parent()->idx, speciestree->root);
	//				if (currnode->parent()->parent()->idx != 0)  // just a test for correctness
	//				{
	//					cout << "ERROR!! root not idx 0 in Euler Tour!" << endl;
	//					exit(1);
	//				}
//					PreOrderSpeciesTree(treeroot, speciestree->root);
					
				}
				
				else  // if moving down to the right-child (i.e., if the current node is the right child of its parent) 
				{
					if(currnode == currnode->parent()->child(1))
					{
						// call the functions for updating the tables for the two nodes that need to be updated (parent and grandparent of the current node)
						PostOrderSpeciesTreeReroot(currnode->parent(), currnode->parent()->getSibling()->idx, currnode->getSibling()->idx, speciestree->root);
//						PreOrderSpeciesTree(currnode->parent(), speciestree->root);
						
						PostOrderSpeciesTreeReroot(treeroot, currnode->parent()->idx, currnode->idx, speciestree->root);
//						PreOrderSpeciesTree(treeroot, speciestree->root);
					}
					else   // sanity check
					{
						cout << "ERROR in Euler Tour!!" << endl;
						exit(1);
					}
				}
					
			}
			
			else  // regular case, when the grand parent of the current node is not the root node of the original gene tree.
			{
				
				// if moving down to the left-child (i.e., if the current node is the left child of its parent) 
				if(currnode == currnode->parent()->child(0))
				{	
					// call the functions for updating the tables for the two nodes that need to be updated (parent and grandparent of the current node)
					PostOrderSpeciesTreeReroot(currnode->parent(), currnode->getSibling()->idx, currnode->parent()->parent()->idx, speciestree->root);
//					PreOrderSpeciesTree(currnode->parent(), speciestree->root);
					
					PostOrderSpeciesTreeReroot(treeroot, currnode->idx, currnode->parent()->idx, speciestree->root);
//					PreOrderSpeciesTree(treeroot, speciestree->root);
					
				}
				
				else  // if moving down to the right-child (i.e., if the current node is the right child of its parent) 
				{
					if(currnode == currnode->parent()->child(1))
					{
						// call the functions for updating the tables for the two nodes that need to be updated (parent and grandparent of the current node)
						PostOrderSpeciesTreeReroot(currnode->parent(), currnode->parent()->parent()->idx, currnode->getSibling()->idx, speciestree->root);
//						PreOrderSpeciesTree(currnode->parent(), speciestree->root);
						
						PostOrderSpeciesTreeReroot(treeroot, currnode->parent()->idx, currnode->idx, speciestree->root);
//						PreOrderSpeciesTree(treeroot, speciestree->root);
					}
					else   // sanity check
					{
						cout << "ERROR in Euler Tour!!" << endl;
						exit(1);
					}
				}
				
					
			}
			
			
			// update best score and best rooting
			
//cout << currnode << " " << outputCurrentCost() << endl;
				if (outputCurrentCost() < bestScore)
				{
					bestScore = outputCurrentCost();
					bestRoot = currnode;
					bestRootQueue.clear();
					temproot.optroot = bestRoot;
					bestRootQueue.push_back(temproot);
					numOptRootings = 1;
				}
				else if (outputCurrentCost() == bestScore)
				{
					numOptRootings++;
					temproot.optroot = currnode;
					bestRootQueue.push_back(temproot);
				}
			}
			

		}

														
//cout << "here 2" << endl;								
			

		if (!currnode->isLeaf())
		{
			eulerTour(currnode->child(0), treeroot);
		
//cout << "here 3" << endl;		
			if (currnode->idx != 0)  // skip if at the root node
			{
				if (currnode->parent()->idx != 0)  // regular case when the current node is not a child of the root in the original gene tree
				{

					// moving up from left-child to current node, and if the parent edge is a left edge in the original gene tree. 
					if(currnode == currnode->parent()->child(0))
					{	
						// call the functions for updating the tables for the two nodes that need to be updated (parent and sibling of the current node)
						PostOrderSpeciesTreeReroot(currnode, currnode->child(0)->idx, currnode->child(1)->idx, speciestree->root);
//						PreOrderSpeciesTree(currnode, speciestree->root);
						
						PostOrderSpeciesTreeReroot(treeroot, currnode->idx, currnode->parent()->idx ,speciestree->root);
//						PreOrderSpeciesTree(treeroot, speciestree->root);
						
					}
					else  // moving up from left-child to current node, and if the parent edge is a right edge in the original gene tree. 
					{
						if(currnode == currnode->parent()->child(1))
						{
							// call the functions for updating the tables for the two nodes that need to be updated (parent and sibling of the current node)
							PostOrderSpeciesTreeReroot(currnode, currnode->child(0)->idx, currnode->child(1)->idx, speciestree->root);
//							PreOrderSpeciesTree(currnode, speciestree->root);
							
							PostOrderSpeciesTreeReroot(treeroot, currnode->parent()->idx, currnode->idx, speciestree->root);
//							PreOrderSpeciesTree(treeroot, speciestree->root);
						}
						else   // sanity check
						{
							cout << "ERROR in Euler Tour!!" << endl;
							exit(1);
						}
					}
				}
				
				else     // special case when the current node is a child of the root in the original gene tree.
				{
					// moving up from left-child to current node, and if the parent edge is a left edge in the original gene tree. 
					if(currnode == currnode->parent()->child(0))
					{	
						// call the functions for updating the tables for the two nodes that need to be updated (parent and sibling of the current node)
						PostOrderSpeciesTreeReroot(currnode, currnode->child(0)->idx, currnode->child(1)->idx, speciestree->root);
//						PreOrderSpeciesTree(currnode, speciestree->root);
						
						PostOrderSpeciesTreeReroot(treeroot, currnode->idx, currnode->getSibling()->idx ,speciestree->root);
//						PreOrderSpeciesTree(treeroot, speciestree->root);
						
					}
					else  // moving up from left-child to current node, and if the parent edge is a right edge in the original gene tree. 
					{
						if(currnode == currnode->parent()->child(1))
						{
							// call the functions for updating the tables for the two nodes that need to be updated (parent and sibling of the current node)
							PostOrderSpeciesTreeReroot(currnode, currnode->child(0)->idx, currnode->child(1)->idx, speciestree->root);
//							PreOrderSpeciesTree(currnode, speciestree->root);
							
							PostOrderSpeciesTreeReroot(treeroot, currnode->getSibling()->idx, currnode->idx, speciestree->root);
//							PreOrderSpeciesTree(treeroot, speciestree->root);
						}
						else   // sanity check
						{
							cout << "ERROR in Euler Tour!!" << endl;
							exit(1);
						}
					}
					
				}
					
			}
//cout << currnode << " " << "up left " << outputCurrentCost() << endl;			
		}


		
		
		
		
		
		if (!currnode->isLeaf())
		{
			eulerTour(currnode->child(1), treeroot);
		
		
			if (currnode->idx != 0)  // skip if at the root node
			{
				if (currnode->parent()->idx != 0)  // regular case when the current node is not a child of the root in the original gene tree
				{

					// moving up from right-child to current node, and if the parent edge is a left edge in the original gene tree. 
					if(currnode == currnode->parent()->child(0))
					{	
						// call the functions for updating the tables for the two nodes that need to be updated (parent and sibling of the current node)
						PostOrderSpeciesTreeReroot(currnode, currnode->child(0)->idx, currnode->child(1)->idx, speciestree->root);
//						PreOrderSpeciesTree(currnode, speciestree->root);
						
						PostOrderSpeciesTreeReroot(treeroot, currnode->idx, currnode->parent()->idx ,speciestree->root);
//						PreOrderSpeciesTree(treeroot, speciestree->root);
						
					}
					else  // moving up from right-child to current node, and if the parent edge is a right edge in the original gene tree. 
					{
						if(currnode == currnode->parent()->child(1))
						{
							// call the functions for updating the tables for the two nodes that need to be updated (parent and sibling of the current node)
							PostOrderSpeciesTreeReroot(currnode, currnode->child(0)->idx, currnode->child(1)->idx, speciestree->root);
//							PreOrderSpeciesTree(currnode, speciestree->root);
							
							PostOrderSpeciesTreeReroot(treeroot, currnode->parent()->idx, currnode->idx, speciestree->root);
//							PreOrderSpeciesTree(treeroot, speciestree->root);
						}
						else   // sanity check
						{
							cout << "ERROR in Euler Tour!!" << endl;
							exit(1);
						}
					}
				}
				
				else     // special case when the current node is a child of the root in the original gene tree.
				{
					// moving up from left-child to current node, and if the parent edge is a left edge in the original gene tree. 
					if(currnode == currnode->parent()->child(0))
					{	
						// call the functions for updating the tables for the two nodes that need to be updated (parent and sibling of the current node)
						PostOrderSpeciesTreeReroot(currnode, currnode->child(0)->idx, currnode->child(1)->idx, speciestree->root);
//						PreOrderSpeciesTree(currnode, speciestree->root);
						
						PostOrderSpeciesTreeReroot(treeroot, currnode->idx, currnode->getSibling()->idx ,speciestree->root);
//						PreOrderSpeciesTree(treeroot, speciestree->root);
						
					}
					else  // moving up from left-child to current node, and if the parent edge is a right edge in the original gene tree. 
					{
						if(currnode == currnode->parent()->child(1))
						{
							// call the functions for updating the tables for the two nodes that need to be updated (parent and sibling of the current node)
							PostOrderSpeciesTreeReroot(currnode, currnode->child(0)->idx, currnode->child(1)->idx, speciestree->root);
//							PreOrderSpeciesTree(currnode, speciestree->root);
							
							PostOrderSpeciesTreeReroot(treeroot, currnode->getSibling()->idx, currnode->idx, speciestree->root);
//							PreOrderSpeciesTree(treeroot, speciestree->root);
						}
						else   // sanity check
						{
							cout << "ERROR in Euler Tour!!" << endl;
							exit(1);
						}
					}
					
				}
					
			}
			
//cout << currnode << " " << "up right " << outputCurrentCost() << endl;			
		}
		
	}

	
	
	
	// output the reconciliation cost for the different rerootings
	int outputCurrentCost()
	{
		
		int minReconCost = INT_MAX / 10;
		int numSpecies_nodes = speciestree->nodes.size();
//		int bestMapIdx;
		
		for (int i = 0; i < numSpecies_nodes; i++)
		{
			if (DParray[0][i] < minReconCost)
			{
				minReconCost = DParray[0][i];
//				bestMapIdx = i;
			}
		}
		return minReconCost;
		
	}
	
	
	// corresponding main procedure for computing costs while rerooting. 
	void PostOrderSpeciesTreeReroot(GeneNodeRooted *currnode, int leftindex, int rightindex, SpeciesNode *CurrSpeciesNode) 
	{
		
		
		if (CurrSpeciesNode == NULL) return;
		PostOrderSpeciesTreeReroot(currnode, leftindex, rightindex, CurrSpeciesNode->child(0));
		PostOrderSpeciesTreeReroot(currnode, leftindex, rightindex, CurrSpeciesNode->child(1));
		
		
		// build up the DP table
		if(currnode->isLeaf() == false)
		{
			
			optQueue.clear(); // clear the queue that stores all optimal solutions
			//			data temp; // create temporary instance of structure for storing optimal mappings and event type.
			//			datainout tempinout; // create temporary instance of inout structure for storing optimal mapping.
			
			if(CurrSpeciesNode->isLeaf() == true)
			{
				
				
				// We always divide INT_MAX by 10 just to prevent overflow of the int value when we are adding two large integers.
				
				minFinal = INT_MAX / 10;  // to keep track of min value among speciation, duplication and transfer.
				
				//speciation 
				c_spec = INT_MAX / 10;
				
				
				// duplication 
				c_dup = DCOST + DParray[leftindex][CurrSpeciesNode->idx] + DParray[rightindex][CurrSpeciesNode->idx];
				
				if (c_dup < minFinal) 
				{
					minFinal = c_dup;
					optQueue.clear();
					temp.mapping0 = CurrSpeciesNode;
					temp.mapping1 = CurrSpeciesNode;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				else if ((c_dup == minFinal) || (minFinal >= INT_MAX / 10)) 
				{
					temp.mapping0 = CurrSpeciesNode;
					temp.mapping1 = CurrSpeciesNode;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				
				
				
				// transfer
				c_trans = INT_MAX / 10;
				int T1, T2;
				if(CurrSpeciesNode != speciestree->root)
				{
					// first transfer scenario
					T1 = in[leftindex][CurrSpeciesNode->idx] + getBestOutReroot(rightindex, CurrSpeciesNode);
					
					c_trans = TCOST + T1;
					
					if (c_trans < minFinal) 
					{
						minFinal = c_trans;
						optQueue.clear();
						temp.mapping0 = inTrack[leftindex][CurrSpeciesNode->idx].optmapping;
						temp.mapping1 = outTrack[rightindex][CurrSpeciesNode->idx].optmapping;
						temp.recipient = temp.mapping1;
						temp.event = 2;
						optQueue.push_back(temp);
					}
					else if ((c_trans == minFinal) || (minFinal >= INT_MAX / 10)) 
					{
						temp.mapping0 = inTrack[leftindex][CurrSpeciesNode->idx].optmapping;
						temp.mapping1 = outTrack[rightindex][CurrSpeciesNode->idx].optmapping;
						temp.recipient = temp.mapping1;
						temp.event = 2;
						optQueue.push_back(temp);
					}
					
					
					
					// second transfer scenario
					T2 = in[rightindex][CurrSpeciesNode->idx] + getBestOutReroot(leftindex, CurrSpeciesNode);
					
					c_trans = TCOST + T2;
					
					if (c_trans < minFinal) 
					{
						minFinal = c_trans;
						optQueue.clear();
						temp.mapping0 = outTrack[leftindex][CurrSpeciesNode->idx].optmapping;
						temp.mapping1 = inTrack[rightindex][CurrSpeciesNode->idx].optmapping;
						temp.recipient = temp.mapping0;
						temp.event = 2;
						optQueue.push_back(temp);
					}
					else if ((c_trans == minFinal) || (minFinal >= INT_MAX / 10)) 
					{
						temp.mapping0 = outTrack[leftindex][CurrSpeciesNode->idx].optmapping;
						temp.mapping1 = inTrack[rightindex][CurrSpeciesNode->idx].optmapping;
						temp.recipient = temp.mapping0;
						temp.event = 2;
						optQueue.push_back(temp);
					}
					
				}
				
				
				
				// pick a randon optimal solution from the queue
				if (!optQueue.empty()) 
				{
					index = rand() % optQueue.size();
				}
				else
				{
					cout << "ERROR!! Empty Queue." << endl;
					exit(1);
				}	
				
				
				// set DParray entry 
				DParray[currnode->idx][CurrSpeciesNode->idx] = minFinal;
				
				// update the backtracking array
				backTrack[currnode->idx][CurrSpeciesNode->idx].mapping0 =  optQueue[index].mapping0;
				backTrack[currnode->idx][CurrSpeciesNode->idx].mapping1 =  optQueue[index].mapping1;
				backTrack[currnode->idx][CurrSpeciesNode->idx].recipient =  optQueue[index].recipient;
				backTrack[currnode->idx][CurrSpeciesNode->idx].event =  optQueue[index].event;
				
				
				
				// update in[][] entry as well as inTrack
				in[currnode->idx][CurrSpeciesNode->idx] = DParray[currnode->idx][CurrSpeciesNode->idx];
				inTrack[currnode->idx][CurrSpeciesNode->idx].optmapping = CurrSpeciesNode;
				
				// update in_second[][] entry as well as in_secondTrack
				in_second[currnode->idx][CurrSpeciesNode->idx] = DParray[currnode->idx][CurrSpeciesNode->idx];
				in_secondTrack[currnode->idx][CurrSpeciesNode->idx].optmapping = CurrSpeciesNode;
				
				
				/*				
				 if((currnode->idx ==1) && (CurrSpeciesNode->idx == 1))
				 cout << endl << "BEGIN " << DParray[currnode->child(0)->idx][CurrSpeciesNode->idx] << "  " <<  out[currnode->child(1)->idx][CurrSpeciesNode->idx] << " " << DParray[currnode->idx][CurrSpeciesNode->idx];
				 */				
			}
			else
			{
				
				minFinal = INT_MAX / 10;  // to keep track of min value among speciation, duplication and transfer.
				
				// speciation
				int S1 = in[leftindex][CurrSpeciesNode->child(0)->idx] + in[rightindex][CurrSpeciesNode->child(1)->idx];
				
				c_spec = S1;
				
				if (c_spec < minFinal) 
				{
					minFinal = c_spec;
					optQueue.clear();
					temp.mapping0 = inTrack[leftindex][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.mapping1 = inTrack[rightindex][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 0;
					optQueue.push_back(temp);
				}
				else if ((c_spec == minFinal) || (minFinal >= INT_MAX / 10)) 
				{
					temp.mapping0 = inTrack[leftindex][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.mapping1 = inTrack[rightindex][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 0;
					optQueue.push_back(temp);
				}
				
				
				
				int S2 = in[leftindex][CurrSpeciesNode->child(1)->idx] + in[rightindex][CurrSpeciesNode->child(0)->idx];
				
				c_spec = S2;
				
				if (c_spec < minFinal) 
				{
					minFinal = c_spec;
					optQueue.clear();
					temp.mapping0 = inTrack[leftindex][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.mapping1 = inTrack[rightindex][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 0;
					optQueue.push_back(temp);
				}
				else if ((c_spec == minFinal) || (minFinal >= INT_MAX / 10)) 
				{
					temp.mapping0 = inTrack[leftindex][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.mapping1 = inTrack[rightindex][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 0;
					optQueue.push_back(temp);
				}
				
				
				
				
				
				
				// duplication
				//				minD = INT_MAX / 10;
				int D1 = LCOST + DParray[leftindex][CurrSpeciesNode->idx] + in[rightindex][CurrSpeciesNode->child(1)->idx];
				
				c_dup = DCOST + D1;
				
				if (c_dup < minFinal) 
				{
					minFinal = c_dup;
					optQueue.clear();
					temp.mapping0 = CurrSpeciesNode;
					temp.mapping1 = inTrack[rightindex][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				else if ((c_dup == minFinal) || (minFinal >= INT_MAX / 10)) 
				{
					temp.mapping0 = CurrSpeciesNode;
					temp.mapping1 = inTrack[rightindex][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				
				
				int D2 = LCOST + DParray[leftindex][CurrSpeciesNode->idx] + in[rightindex][CurrSpeciesNode->child(0)->idx];
				
				c_dup = DCOST + D2;
				
				if (c_dup < minFinal) 
				{
					minFinal = c_dup;
					optQueue.clear();
					temp.mapping0 = CurrSpeciesNode;
					temp.mapping1 = inTrack[rightindex][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				else if ((c_dup == minFinal) || (minFinal >= INT_MAX / 10)) 
				{
					temp.mapping0 = CurrSpeciesNode;
					temp.mapping1 = inTrack[rightindex][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				
				int D3 = LCOST + DParray[rightindex][CurrSpeciesNode->idx] + in[leftindex][CurrSpeciesNode->child(1)->idx];
				
				c_dup = DCOST + D3;
				
				if (c_dup < minFinal) 
				{
					minFinal = c_dup;
					optQueue.clear();
					temp.mapping0 = inTrack[leftindex][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.mapping1 = CurrSpeciesNode;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				else if ((c_dup == minFinal) || (minFinal >= INT_MAX / 10)) 
				{
					temp.mapping0 = inTrack[leftindex][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.mapping1 = CurrSpeciesNode;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				
				int D4 = LCOST + DParray[rightindex][CurrSpeciesNode->idx] + in[leftindex][CurrSpeciesNode->child(0)->idx];
				
				c_dup = DCOST + D4;
				
				if (c_dup < minFinal) 
				{
					minFinal = c_dup;
					optQueue.clear();
					temp.mapping0 = inTrack[leftindex][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.mapping1 = CurrSpeciesNode;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				else if ((c_dup == minFinal) || (minFinal >= INT_MAX / 10)) 
				{
					temp.mapping0 = inTrack[leftindex][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.mapping1 = CurrSpeciesNode;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				
				int D5 = DParray[leftindex][CurrSpeciesNode->idx] + DParray[rightindex][CurrSpeciesNode->idx];
				
				c_dup = DCOST + D5;
				
				if (c_dup < minFinal) 
				{
					minFinal = c_dup;
					optQueue.clear();
					temp.mapping0 = CurrSpeciesNode;
					temp.mapping1 = CurrSpeciesNode;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				else if ((c_dup == minFinal) || (minFinal >= INT_MAX / 10)) 
				{
					temp.mapping0 = CurrSpeciesNode;
					temp.mapping1 = CurrSpeciesNode;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				
				
				int D6 = (2 * LCOST) + in[leftindex][CurrSpeciesNode->child(0)->idx] + in[rightindex][CurrSpeciesNode->child(1)->idx];
				
				c_dup = DCOST + D6;
				
				if (c_dup < minFinal) 
				{
					minFinal = c_dup;
					optQueue.clear();
					temp.mapping0 = inTrack[leftindex][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.mapping1 = inTrack[rightindex][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				else if ((c_dup == minFinal) || (minFinal >= INT_MAX / 10)) 
				{
					temp.mapping0 = inTrack[leftindex][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.mapping1 = inTrack[rightindex][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				
				int D7 = (2 * LCOST) + in[leftindex][CurrSpeciesNode->child(1)->idx] + in[rightindex][CurrSpeciesNode->child(0)->idx];
				
				c_dup = DCOST + D7;
				
				if (c_dup < minFinal) 
				{
					minFinal = c_dup;
					optQueue.clear();
					temp.mapping0 = inTrack[leftindex][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.mapping1 = inTrack[rightindex][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				else if ((c_dup == minFinal) || (minFinal >= INT_MAX / 10)) 
				{
					temp.mapping0 = inTrack[leftindex][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.mapping1 = inTrack[rightindex][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				
				int D8 = (2 * LCOST) + in[leftindex][CurrSpeciesNode->child(0)->idx] + in[rightindex][CurrSpeciesNode->child(0)->idx];
				
				c_dup = DCOST + D8;
				
				if (c_dup < minFinal) 
				{
					minFinal = c_dup;
					optQueue.clear();
					temp.mapping0 = inTrack[leftindex][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.mapping1 = inTrack[rightindex][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				else if ((c_dup == minFinal) || (minFinal >= INT_MAX / 10)) 
				{
					temp.mapping0 = inTrack[leftindex][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.mapping1 = inTrack[rightindex][CurrSpeciesNode->child(0)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				
				int D9 = (2 * LCOST) + in[leftindex][CurrSpeciesNode->child(1)->idx] + in[rightindex][CurrSpeciesNode->child(1)->idx];
				
				c_dup = DCOST + D9;
				
				if (c_dup < minFinal) 
				{
					minFinal = c_dup;
					optQueue.clear();
					temp.mapping0 = inTrack[leftindex][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.mapping1 = inTrack[rightindex][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				else if ((c_dup == minFinal) || (minFinal >= INT_MAX / 10)) 
				{
					temp.mapping0 = inTrack[leftindex][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.mapping1 = inTrack[rightindex][CurrSpeciesNode->child(1)->idx].optmapping;
					temp.recipient = NULL;
					temp.event = 1;
					optQueue.push_back(temp);
				}
				
				
				
				
				// transfer
				c_trans = INT_MAX / 10; 
				if(CurrSpeciesNode != speciestree->root)
				{
					int T1 = in[leftindex][CurrSpeciesNode->idx] + getBestOutReroot(rightindex, CurrSpeciesNode);
					
					
					c_trans = TCOST + T1;
					
					if (c_trans < minFinal) 
					{
						minFinal = c_trans;
						optQueue.clear();
						temp.mapping0 = inTrack[leftindex][CurrSpeciesNode->idx].optmapping;
						temp.mapping1 = outTrack[rightindex][CurrSpeciesNode->idx].optmapping;
						temp.recipient = temp.mapping1;
						temp.event = 2;
						optQueue.push_back(temp);
					}
					else if ((c_trans == minFinal) || (minFinal >= INT_MAX / 10)) 
					{
						temp.mapping0 = inTrack[leftindex][CurrSpeciesNode->idx].optmapping;
						temp.mapping1 = outTrack[rightindex][CurrSpeciesNode->idx].optmapping;
						temp.recipient = temp.mapping1;
						temp.event = 2;
						optQueue.push_back(temp);
					}
					
					
					int T2 = in[rightindex][CurrSpeciesNode->idx] + getBestOutReroot(leftindex, CurrSpeciesNode);
					
					c_trans = TCOST + T2;
					
					if (c_trans < minFinal) 
					{
						minFinal = c_trans;
						optQueue.clear();
						temp.mapping0 = outTrack[leftindex][CurrSpeciesNode->idx].optmapping;
						temp.mapping1 = inTrack[rightindex][CurrSpeciesNode->idx].optmapping;
						temp.recipient = temp.mapping0;
						temp.event = 2;
						optQueue.push_back(temp);
					}
					else if ((c_trans == minFinal) || (minFinal >= INT_MAX / 10)) 
					{
						temp.mapping0 = outTrack[leftindex][CurrSpeciesNode->idx].optmapping;
						temp.mapping1 = inTrack[rightindex][CurrSpeciesNode->idx].optmapping;
						temp.recipient = temp.mapping0;
						temp.event = 2;
						optQueue.push_back(temp);
					}
					
				}
				
				
				
				// pick a random optimal solution from the queue
				if (!optQueue.empty()) 
				{
					index = rand() % optQueue.size();
				}
				else
				{
					cout << "ERROR!! Empty Queue." << endl;
					exit(1);
				}
				
				
				// set DParray entry
				DParray[currnode->idx][CurrSpeciesNode->idx] = minFinal;
				
				// update the backtracking array
				backTrack[currnode->idx][CurrSpeciesNode->idx].mapping0 =  optQueue[index].mapping0;
				backTrack[currnode->idx][CurrSpeciesNode->idx].mapping1 =  optQueue[index].mapping1;
				backTrack[currnode->idx][CurrSpeciesNode->idx].recipient =  optQueue[index].recipient;
				backTrack[currnode->idx][CurrSpeciesNode->idx].event =  optQueue[index].event;
				
				
				// update in[][] entry and the inTrack table
				
				minIn = DParray[currnode->idx][CurrSpeciesNode->idx];
				tempinout.optmapping = CurrSpeciesNode; 
				if ( (LCOST + in[currnode->idx][CurrSpeciesNode->child(0)->idx]) < minIn)
				{
					minIn = (LCOST + in[currnode->idx][CurrSpeciesNode->child(0)->idx]);
					tempinout.optmapping = inTrack[currnode->idx][CurrSpeciesNode->child(0)->idx].optmapping;
				}
				if ( (LCOST + in[currnode->idx][CurrSpeciesNode->child(1)->idx]) < minIn)
				{
					minIn = (LCOST + in[currnode->idx][CurrSpeciesNode->child(1)->idx]);
					tempinout.optmapping = inTrack[currnode->idx][CurrSpeciesNode->child(1)->idx].optmapping;
				}
				
				in[currnode->idx][CurrSpeciesNode->idx] = minIn;
				inTrack[currnode->idx][CurrSpeciesNode->idx].optmapping = tempinout.optmapping;
				
				
				
				
				// update in_second[][] entry and the in_secondTrack table
				
				minIn_second = DParray[currnode->idx][CurrSpeciesNode->idx];
				tempinout.optmapping = CurrSpeciesNode; 
				if ( in_second[currnode->idx][CurrSpeciesNode->child(0)->idx] < minIn_second)
				{
					minIn_second = in_second[currnode->idx][CurrSpeciesNode->child(0)->idx];
					tempinout.optmapping = in_secondTrack[currnode->idx][CurrSpeciesNode->child(0)->idx].optmapping;
				}
				if ( in_second[currnode->idx][CurrSpeciesNode->child(1)->idx] < minIn_second)
				{
					minIn_second = in_second[currnode->idx][CurrSpeciesNode->child(1)->idx];
					tempinout.optmapping = in_secondTrack[currnode->idx][CurrSpeciesNode->child(1)->idx].optmapping;
				}
				
				in_second[currnode->idx][CurrSpeciesNode->idx] = minIn_second;
				in_secondTrack[currnode->idx][CurrSpeciesNode->idx].optmapping = tempinout.optmapping;
				
			}
			
		}
	}
	
	
	
	
	void runForNewRoot(int i2, ostream &os, bool random)
	{
		GeneTreeRooted &tree = *genetree_rooted[i2];
		tree.assignIndex();
		
		
		// clean up the arrays
		for (int i =0 ; i< DParray.size(); i++)
		{
			DParray[i].clear();
			in[i].clear();
			in_second[i].clear();
			out[i].clear();
			
			backTrack[i].clear();
			inTrack[i].clear();
			outTrack[i].clear();
			in_secondTrack[i].clear();
			
		}
		DParray.clear();
		in.clear();
		in_second.clear();
		out.clear();
		
		backTrack.clear();
		inTrack.clear();
		outTrack.clear();
		in_secondTrack.clear();
		
		
		
		// initialize the numDup and numTrans counters (to store the final number of duplications and transfers in the reconciliation)
		numDup = 0;
		numTrans = 0;
		
		// Initialize the dynamic programming table and in and out counters to infinity
		
		int numGene_nodes = tree.nodes.size();
		int numSpecies_nodes = speciestree->nodes.size();
		
		vector<int> row;
		row.clear();
		
		vector<data> datarow;
		datarow.clear();
		data temp1;
		vector<datainout> datainoutrow;
		datainoutrow.clear();
		datainout temp2;
		
		
		for (int i = 0; i< numGene_nodes; i++)
		{		
			
			for (int j = 0; j< numSpecies_nodes;j++)
			{
				row.push_back(INT_MAX / 10);
				
				temp1.mapping0 = NULL;
				temp1.mapping1 = NULL;
				temp1.recipient = NULL;
				temp1.event = -1;
				datarow.push_back(temp1);
				
				temp2.optmapping = NULL;
				datainoutrow.push_back(temp2);
				
			}
			
			DParray.push_back(row);
			in.push_back(row);
			out.push_back(row);
			in_second.push_back(row);
			
			backTrack.push_back(datarow);
			inTrack.push_back(datainoutrow);
			outTrack.push_back(datainoutrow);
			in_secondTrack.push_back(datainoutrow);
			
			row.clear();
			datarow.clear();
			datainoutrow.clear();
		}
		
		
		
		
		// base case of the dynamic programming table (incorporate leaf to leaf mapping)	
		vector<GeneNodeRooted*> &g = tree.nodes;
		SpeciesNode *temp_node;
		
		
		
		for(vector<GeneNodeRooted*>::iterator itr=g.begin(); itr!=g.end(); itr++) 
		{
			GeneNodeRooted *genenode = *itr;
			if (genenode->isLeaf())
			{
				SpeciesNode *&mapping = genenode->getMapping();
				DParray[genenode->idx][mapping->idx] = 0;
				
				temp_node = genenode->getMapping();
				int count = 0;
				while (1)
				{
					in[genenode->idx][temp_node->idx] = (count * LCOST);
					in_second[genenode->idx][temp_node->idx] = 0;
					
					inTrack[genenode->idx][temp_node->idx].optmapping = genenode->getMapping();
					in_secondTrack[genenode->idx][temp_node->idx].optmapping = genenode->getMapping();
					
					if (temp_node == speciestree->root)
						break;
					temp_node = temp_node->parent();
					count++;
					
				}
				
			}
		}
		
		PostOrderGeneTree(tree.root);
		
			// Output the species tree and gene tree
		if(random ==1)
		{	
			os << endl << endl << " ------------ An Optimal Rooting for Gene Tree " << i2+1 << " (Unrooted) -------------" << endl;
//			os << "Species Tree: " << endl; 
//			speciestree->tree2newick(os);
//			os << endl << endl;
//			os << "Optimally rooted Gene Tree: " << endl;
			tree.genetree2newick(os);
			os << endl;
		}
			//	cout << tree << endl;
		
		// decorate the nodes of the gene tree with their leftmost and rightmost descendants.
		leftRightDescendants(tree.root);
		
		// output the final reconciliation cost
		OutputReconciliationCost(os,tree.root, random);
		
	}
		
	
	
	
	
	
	
	
	
	
			// This function computes the best receiver for a transfer (by brute force to account for distance dependant transfer costs).
	int getBestOutReroot(int currnodeIndex, SpeciesNode *node)
	{
		minOut = INT_MAX / 10;
			
		bool valid;
		int transferAdd;   // additional cost of transfer based on distance
		int num_nodes = speciestree->nodes.size();
		int distance;
		SpeciesNode* temp_node;
		SpeciesNode* lcanode;
		
		// iterate over all non-root nodes
		for (int i = 1; i< num_nodes; i++)
		{
			
//if (speciestree->nodes[i]->child(0) == NULL)
//cout << speciestree->nodes[i]->branchDepth << endl;
			
                       if (speciestree->nodes[i] == node)
                                continue;


                        valid = true;

                        // disallow transfers that go from descendant to ancestor or vice versa. All other transfers are OK.
                        if ((node->no >= speciestree->nodes[i]->begin) && (node->no <= speciestree->nodes[i]->end))
                                valid = false;
                        if ((speciestree->nodes[i]->no >= node->begin) && (speciestree->nodes[i]->no <= node->end))
                                valid = false;

			
			
			if (valid == true)
			{
				
				temp_node = node;
				while(temp_node->parent() != NULL)
				{
					temp_node = temp_node->parent();
					if((speciestree->nodes[i]->no >= temp_node->begin) && (speciestree->nodes[i]->no <= temp_node->end))
						break;
				}
				
				lcanode = temp_node;
				
				
				distance = (node->nodeDepth - lcanode->nodeDepth) + (speciestree->nodes[i]->nodeDepth - lcanode->nodeDepth);
				
				
				
				
				
//				if((node->no >= speciestree->nodes[i]->begin) && (node->no <= speciestree->nodes[i]->end))
//					cout << "ERROR!!!" << endl;
//				if((speciestree->nodes[i]->no >= node->begin) && (speciestree->nodes[i]->no <= node->end))
//					cout << "ERROR!!!" << endl;
				
				
				if (TYPE == 0)   // if not using variable transfer costs
				{
					if (DParray[currnodeIndex][speciestree->nodes[i]->idx] < minOut)
					{
						minOut = DParray[currnodeIndex][speciestree->nodes[i]->idx];
						outTrack[currnodeIndex][node->idx].optmapping = speciestree->nodes[i];
					}
				}
				if (TYPE == 1)  // if using the simple transfer cost
				{
					
					if (distance >= THRESHOLD)
						transferAdd = ADD;
					else
						transferAdd = 0;
					
					if ((DParray[currnodeIndex][speciestree->nodes[i]->idx] + transferAdd) < minOut)
					{
						minOut = (DParray[currnodeIndex][speciestree->nodes[i]->idx] + transferAdd);
						outTrack[currnodeIndex][node->idx].optmapping = speciestree->nodes[i];
					}
					
				}
				
				
				if (TYPE == 2)   // if using the more general transfer cost
				{
					
					if (distance >= THRESHOLD)
						transferAdd = (distance/THRESHOLD) * ADD;
					else
						transferAdd = 0;
					
					if ((DParray[currnodeIndex][speciestree->nodes[i]->idx] + transferAdd) < minOut)
					{
						minOut = (DParray[currnodeIndex][speciestree->nodes[i]->idx] + transferAdd);
						outTrack[currnodeIndex][node->idx].optmapping = speciestree->nodes[i];
					}
				}
				
				

				
				
			}

		}
 		
		return minOut;
	}
	
	
	
	
	
};


#endif
